#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# This file is part of DAWIS 3.0.
#
# NAME : hard_threshold.py
# AUTHOR : Amael Ellien
# LAST MODIFICATION : 07/2021
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# Hard thresholding process for the multiscale support. The threshold is a
# simple constant given by N_SIGMA*SIGMA.
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

from dawis.table_noise_stds import table_noise_stds
from dawis.datacube import *
from scipy.stats import sigmaclip
from scipy.stats import normaltest
import numpy as np
import matplotlib.pyplot as plt
import logging

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def hard_threshold(wavelet_datacube, n_sigmas = 3, wavelet_type = None, keep_neg = False, verbose = False, display = False):

    '''
    Hard thresholding process for the multiscale support. The threshold is a simple constant given by N_SIGMA*SIGMA.
    '''

    # Add exceptions

    x_axis, y_axis, n_levels = wavelet_datacube.shape

    support = np.zeros((x_axis, y_axis, n_levels))

    if wavelet_type:
        conv_table = table_noise_stds(n_levels, wavelet_type = wavelet_type)
    else:
        wavelet_type = wavelet_datacube.fheader['WAVTYPE']
        conv_table = table_noise_stds(n_levels, wavelet_type = wavelet_type)

    if verbose == True:
        log = logging.getLogger(__name__)
        log.info('\ncreating support datacube : hard thresholding')

    noise_pixels = np.copy(wavelet_datacube.array[:,:,0]).flatten()
    noise_pixels = np.delete(noise_pixels, np.where(noise_pixels == 0) )
    noise_pixels = sigmaclip(noise_pixels, low = n_sigmas, high = n_sigmas)[0]
    sigma = np.std(noise_pixels)

    alpha = 1E-3
    if noise_pixels[0].size > 0:
        k, pval = normaltest( noise_pixels )
        if pval > alpha:
            log.info('/!\ Warning /!\ noise does not look Gaussian : p = %1.5f alpha = %1.5f'%(pval, alpha))

    if display == True:
        plt.figure(1)
        plt.hist(noise_pixels[0].flatten(),  bins = 'auto',  histtype = 'bar', rwidth = 10. , align='left')
        plt.show()

    for level in range(0, n_levels):

        if keep_neg == True:
            non_masked_pixels = np.where( abs( wavelet_datacube.array[:,:, level] )\
                         >= (n_sigmas * sigma * conv_table[level] ))

        elif keep_neg == False:
            non_masked_pixels = np.where( wavelet_datacube.array[:,:,level]\
                         >= (n_sigmas * sigma * conv_table[level] ))

        temp = support[:,:,level]
        temp[non_masked_pixels] = 1.0
        support[:,:,level] = temp

        if verbose == True:
            log.info('level %02d sigma %f' %(level+1, n_sigmas*sigma*conv_table[level] ))

    support_dc = support_datacube(support, threshold_type = 'HARD', \
                                         noise_pixels = noise_pixels, \
                                         fheader = wavelet_datacube.fheader)
    return support_dc

if __name__ == '__main__':

    #wdc = datacube.from_fits('/home/ellien/devd/tests/cat_gal_004_z_0.1_Ficl_0.4_Re_050_noise_megacam_dcw.fits')
    #wdc.fheader['WAVTYPE'] = 'LBSPL'
    #wdc.to_fits('/home/ellien/devd/tests/cat_gal_004_z_0.1_Ficl_0.4_Re_050_noise_megacam_dcw.fits', overwrite = True)

    from atrous import atrous
    from numpy.random import normal
    from astropy.io import fits
    from wavelet_transforms import bspl_atrous, wavelet_datacube

    hdu = fits.open('/home/ellien/wavelets/A1365.rebin.fits')
    im = hdu[0].data
    header = hdu[0].header
    cdc, wdc = bspl_atrous(im, 10, header)
    cdc.to_fits('/home/ellien/devd/tests/A1365.rebin.cdc.fits', overwrite = True)
    wdc.to_fits('/home/ellien/devd/tests/A1365.rebin.wdc.fits', overwrite = True)
    #wdc.std_plot(name = 'test', markerstyle = 'o', color = 'black', linestyle = '-')
    wdc = wavelet_datacube.from_fits('/home/ellien/devd/tests/A1365.rebin.wdc.fits')
    sdc = hard_threshold(wdc, n_sigmas = 5)
    #sdc.waveplot(cmap = 'binary')
    #sdc.histogram_noise()
    #sdc.test_gaussianity()
    wdc.fheader



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# This file is part of DAWIS (Detection Algorithm for Intracluster light Studies).
# Author: Amaël Ellien
# Last modification: 02/07/2021
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# MODULES

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import SymLogNorm
from astropy.io import fits
from dawis.exceptions import DawisDimensionError
from dawis.datacube import *
from dawis.atrous import atrous, boundary_conditions, interlaced_atrous_congrid
from dawis.exceptions import DawisWrongType
import pdb
import logging

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def bspl_atrous(image, n_levels, fheader = None, conditions = 'loop', verbose = False):
    '''doc to do'''

    filter = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
    coarse_array, wavelet_array = atrous(image = image, n_levels = n_levels, filter = filter, conditions = 'loop')

    coarse_dc = coarse_datacube(array = coarse_array, wavelet_type = 'BSPL', fheader = fheader)
    wavelet_dc = wavelet_datacube(array = wavelet_array, wavelet_type = 'BSPL', fheader = fheader)

    return coarse_dc, wavelet_dc

def haar_atrous(image, n_levels, fheader = None, verbose = False):
    '''doc to do'''

    filter = 1 / 2. * np.array([ 1, 0, 1 ])
    coarse_array, wavelet_array = atrous(image = image, n_levels = n_levels, filter = filter, conditions = 'loop')

    coarse_dc = coarse_datacube(array = coarse_array, wavelet_type = 'HAAR', fheader = fheader)
    wavelet_dc = wavelet_datacube(array = wavelet_array, wavelet_type = 'HAAR', fheader = fheader)

    return coarse_dc, wavelet_dc

def icbspl_atrous(image, n_levels, n_voices, fheader = None, verbose = False):
    '''doc to do
       For now scale 2^0.5 is always removed from result.'''

    filter = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
    coarse_array, wavelet_array = interlaced_atrous_congrid(image = image, n_levels = n_levels, n_voices = n_voices, filter = filter)
    coarse_array = np.delete(coarse_array, obj = 1, axis = 2 ) # delete coarse scale 2^0.5
    wavelet_array = np.delete(wavelet_array, obj = 0, axis = 2 ) # delete coarse scale 2^0.5
    wavelet_array[:,:,0] = coarse_array[:,:,0] - coarse_array[:,:,1]

    coarse_dc = coarse_datacube(array = coarse_array, wavelet_type = 'BSPL', fheader = fheader)
    wavelet_dc = wavelet_datacube(array = wavelet_array, wavelet_type = 'BSPL', fheader = fheader)

    return coarse_dc, wavelet_dc


if __name__ == '__main__':

    from atrous import atrous
    from numpy.random import normal
    from astropy.io import fits

    im = normal(0, 1,(1024, 1024))
    hdu = fits.open('/home/ellien/wavelets/A1365.rebin.fits')
    im = hdu[0].data
    header = hdu[0].header
    cdc, wdc = haar_atrous(im, 10, header)

    wdc.waveplot(cmap = 'hot')
    cdc.waveplot()



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# This file is part of DAWIS (Detection Algorithm for Intracluster light Studies).
# Author: Amaël Ellien
# Last modification: 27/11/2020
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# MODULES

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class DawisError(Exception):
    """Base class for errors in DAWIS.
    """
    def __repr__(self):
        return 'dawis.DawisError(%r)' %(str(self))


class DawisDimensionError(DawisError):
    """A DAWIS-specific error class raised when the number of dimensions of an
    input array does not match the number of dimension required by the algorithm.
    """
    def __init__(self, message, ndim, allowed_ndim):
        self.message      = message
        self.ndim         = ndim
        self.allowed_ndim = allowed_ndim

        message = ''.join( ( self.message, \
                             ' number of dimensions is ', \
                             str(ndim), \
                             ' instead of ', \
                             str(allowed_ndim) ) )
        super().__init__(message)

    def __repr__(self):
        return 'dawis.DawisDimensionError(%r)' %(str(self))


class DawisUnknownType(DawisError):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message, type):
        self.message = message
        self.type = type

        message = ''.join( ( self.message, \
                             ' type ', \
                             str(type), \
                             ' is unknown.' ))
        super().__init__(message)


class DawisWrongType(DawisError):
    """A DAWIS-specific error class raised when a wrong argument type is given
    to a function or a class.
    """

    def __init__(self, message, type_error, allowed_type):
        self.message = message
        self.type_error = type_error
        self.allowed_type = allowed_type

        message = ''.join( ( self.message, \
                             ' type is ', \
                             str(type_error), \
                             ' instead of ', \
                             str(allowed_type) ) )
        super().__init__(message)

    def __repr__(self):
        return 'dawis.DawisDimensionError(%r)' %(str(self))

class DawisValueError(DawisError):
    """A DAWIS-specific error class raised when a wrong argument value is given
    to a function or a class.
    """

    def __init__(self, message, value_error, allowed_value):
        self.message = message
        self.value_error = value_error
        self.allowed_value = allowed_value

        message = ''.join( ( self.message, \
                             ' value ', \
                             str(value_error), \
                             ' is not allowed, must be ', \
                             str(allowed_value) ) )
        super().__init__(message)

    def __repr__(self):
        return 'dawis.DawisDimensionError(%r)' %(str(self))

if __name__ == '__main__':

    raise DawisDimensionError('Not a datacube -->', 2, 3)
    raise DawisDimensionError('Not an image -->', 3, 2)
